// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#ifndef ICMPV4_PACKET_H
#define ICMPV4_PACKET_H

#include <stdint.h>

#include <istream>
#include <ostream>

#include <boost/asio.hpp>

#include "icmp/icmppacket.h"
#include "icmp/icmpv4header.h"
#include "icmp/icmpdata.h"
#include "icmp/icmptype.h"
#include "ip/ipv4header.h"

//-----------------------------------------------------------------------------
// Icmpv4Packet
//-----------------------------------------------------------------------------

/**
 * @brief This class represents the ICMP version 4 Packet.
 *
 * The ICMP Packet format is:
 *
 * @code
 * 0               8               16                             31
 * +-------+-------+---------------+------------------------------+      ---
 * |       |       |               |                              |       ^
 * |version|header |    type of    |    total length in bytes     |       |
 * |  (4)  | length|    service    |                              |       |
 * +-------+-------+---------------+-+-+-+------------------------+       |
 * |                               | | | |                        |       |
 * |        identification         |0|D|M|    fragment offset     |       |
 * |                               | |F|F|                        |       |
 * +---------------+---------------+-+-+-+------------------------+       |
 * |               |               |                              |       |
 * | time to live  |   protocol    |       header checksum        |  IPv4 Header
 * |               |               |                              |   20 bytes
 * +---------------+---------------+------------------------------+       |
 * |                                                              |       |
 * |                      source IPv4 address                     |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                   destination IPv4 address                   |       |
 * |                                                              |       v
 * +---------------+---------------+------------------------------+      ---
 * |               |               |                              |       ^
 * |     type      |     code      |          checksum            |       |
 * |               |               |                              |       |
 * +---------------+---------------+------------------------------+       |
 * |                               |                              | ICMP Payload
 * |          identifier           |       sequence number        | (header +
 * |                               |                              |       data)
 * +-------------------------------+------------------------------+   8+ bytes
 * |                                                              |       |
 * |                      data (optional)                         |       |
 * |                                                              |       v
 * +-------------------------------+------------------------------+      ---
 * @endcode
 */
class Icmpv4Packet : public IcmpPacket
{
public:
    Icmpv4Packet();
    Icmpv4Packet(
            const Icmpv4Header &icmp_header,
            const IcmpData &icmp_data
    );
    virtual ~Icmpv4Packet();

    Ipv4Header get_ip_header() const;
    Icmpv4Header get_icmp_header() const;
    IcmpData get_icmp_data() const;

    virtual bool match_echo_reply(
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;
    virtual bool match_destination_unreachable(
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_echo_reply(
            const std::size_t &bytes_transferred,
            const boost::posix_time::ptime &time_packet_sent
    ) const;
    virtual void print_destination_unreachable() const;

    virtual ReadReturnCode read( std::istream &is );
    bool write( std::ostream &os ) const;


private:
    bool match(
            const Icmpv4Type type,
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;

private:
    /// The IP header.
    Ipv4Header IpHeader;
    /// The ICMP packet header.
    Icmpv4Header IcmpPayloadHeader;
    /// The ICMP packet payload (data).
    IcmpData IcmpPayloadData;
};

#endif // ICMPV4_PACKET_H
